
// Reference to AES c-model
extern void aescmodel(input bit [127:0], input bit [127:0], input bit, output bit [127:0]);


program automatic test (wb_if.TST wbif);
	
	event handshake;  	// Handshake event for BFM and Generator interaction
	const int ROUNDS=10;	// Number of tests



	

	// *** TRANSACTION CLASS
	// *** Objects for data exchange between transactors
	// ***
	class Transaction;
		rand bit [127:0] data_in;	
		rand bit [127:0] key;
		bit [127:0] data_out;   	// !! comment why this one can be a bit (x-assert is not done on this signal)
		bit [127:0] data_c_out;   	 
		bit decrypt;
		bit reset = 0;
	endclass

	

	// *** GENERATOR CLASS
	// ***
	class Generator;

		
		mailbox gen2bfm;	// Mailbox to BFM
		mailbox gen2scb; 	// Mailbox to Scoreboard
		
		function new(mailbox gen2bfm, gen2scb);
			this.gen2bfm = gen2bfm;
			this.gen2scb = gen2scb;
		endfunction
		
		task run;
			begin
				// Start with one reset to make the AES module work
				reset_task();
			
			
				for (int i=0; i<ROUNDS; i++) begin
			
					// Start new encryption, decryption or reset
					randsequence (stream)
					stream : encrypt := 5 |
						 decrypt := 5 |
						 reset := 5;
					encrypt : { encrypt_task; } |
						  { encrypt_task; } encrypt;
					decrypt : { decrypt_task; } |
						  { decrypt_task; } decrypt;
					reset :	  { reset_task; } |
						  { reset_task; } reset;
					endsequence
				
					//wait for BFM to fetch object
					@handshake;
				end
			end
		endtask
		
		
		// Create tr object for encryption
		task encrypt_task;
			begin
				// new tr object for data storage
				Transaction tr = new();

				// randomize key and indata
				a_randomize_e: assert(tr.randomize());
				tr.decrypt = 0;

				// send tr object to BFM and Scoreboard mailboxes
				gen2bfm.put(tr);
				gen2scb.put(tr);
			end
		endtask
		
		// Create tr object for decryption
		task decrypt_task;
			begin
				// new tr object for data storage
				Transaction tr = new();

				// randomize key and indata
				a_randomize_d: assert(tr.randomize());
				tr.decrypt = 1;

				// send tr object to BFM and Scoreboard mailboxes
				gen2bfm.put(tr);
				gen2scb.put(tr);
			end
		endtask
		
		// Create tr object for reset
		task reset_task;
			begin
				// new tr object for data storage
				Transaction tr = new();

				tr.reset = 1;

				// send tr object to BFM and Scoreboard mailboxes
				gen2bfm.put(tr);
				gen2scb.put(tr);
			end
		endtask
		
		
	endclass
		


	// *** SCOREBOARD CLASS
	// ***
	class Scoreboard;
	
		Transaction tr;

		mailbox gen2scb, scb2chk;

		function new(mailbox gen2scb, scb2chk);
			this.gen2scb = gen2scb;
			this.scb2chk = scb2chk;
		endfunction
		
		task run;
			repeat (ROUNDS) begin
				gen2scb.get(tr);
				
				if (!tr.reset) aescmodel(tr.data_in,tr.key,tr.decrypt,tr.data_c_out);
				scb2chk.put(tr);
			end
		endtask
	endclass	
		

	// *** BFM CALLBACK CLASS
	// ***
	class BFM_cbs;
		task pre_tx();
			repeat ($urandom_range(0,9)) begin
				@(posedge wbif.clk);
			end
		endtask
		
		task post_tx();
			// do nothing for now
		endtask
	endclass		



	// *** BUS FUNCTIONAL MODEL CLASS
	// *** Wishbone bus communication with device-under-test
	// ***
	class BFM;
		int data_cycle_out[4];		//Outdata from Wishbone cycles
		int data_cycle_in[9];		//Indata for WB cycles

		BFM_cbs cbs;
		Transaction tr;	
		
		
		// Define cover group for WB indata
		covergroup covWbDatIn;
			coverpoint wb_if.wb_dat_i;	
		endgroup
		
		
		mailbox gen2bfm;	//Mailbox from Generator
		mailbox bfm2chk;	//Mailbox to Checker
	
		function new(mailbox gen2bfm, bfm2chk);
			this.gen2bfm = gen2bfm;
			this.bfm2chk = bfm2chk;
			covWbDatIn = new;
		endfunction
		

		task run;
			repeat (ROUNDS) begin
				// Get data from Generator
				gen2bfm.get(tr);
				
				// Tell Generator to continue
				->handshake;
				
				
				// Fill array with data from received tr object
				data_cycle_in[8] = tr.key[31:0];
				data_cycle_in[7] = tr.key[63:32];
				data_cycle_in[6] = tr.key[95:64];
				data_cycle_in[5] = tr.key[127:96];
				data_cycle_in[4] = tr.data_in[31:0];
				data_cycle_in[3] = tr.data_in[63:32];
				data_cycle_in[2] = tr.data_in[95:64];
				data_cycle_in[1] = tr.data_in[127:96];
				data_cycle_in[0] = (tr.decrypt*4)+1;	//decrypt: send 3'b101
									//encrypt: send 3'b001
				

				
				if (tr.reset == 1) begin
					// Reset device
					@(posedge wbif.clk);
					sendreset();
					$display("RESET!");				
				end
				else begin

					// Send 9 cycles of data
					for (int i=8; i>=0; i--) begin

						@(posedge wbif.clk);
						writecycle((i*4),data_cycle_in[i]);
									
					end // for		
		
	
					// Wait for AES calculation to complete
					repeat(600) @(posedge wbif.clk);
		

					// Read outdata using 4 cycles				
					for (int i=3; i>=0; i--) begin

						@(posedge wbif.clk);
						readcycle((i*4)+36);

					end // for		


					// Add out data to tr object
					tr.data_out[31:0] = data_cycle_out[3];
					tr.data_out[63:32] = data_cycle_out[2];
					tr.data_out[95:64] = data_cycle_out[1];
					tr.data_out[127:96] = data_cycle_out[0];
				end	
				
				// Send tr object to Checker mailbox
				bfm2chk.put(tr);
		
			end
		endtask
		
		
		// Reset device
		task sendreset();
		begin
			@(posedge wbif.clk);	
			wbif.reset = 1'b0;
			@(posedge wbif.clk);
			wbif.reset = 1'b1;
			@(posedge wbif.clk);
			wbif.reset = 1'b0;
			wbif.wb_cyc = 1'b0;
		end
		endtask
		
		
		// Set signals on Wishbone bus
		task setsignals(bit we, bit [31:0] adr, dat_i);
		begin
			cbs.pre_tx();		// Pre-transfer callback
			
			wbif.wb_adr = adr;	// data address
			wbif.wb_dat_i = dat_i;	// data in
			wbif.wb_cyc = 1'b1;	// cycle indicator
			wbif.wb_stb = 1'b1;	// strobe
			wbif.wb_we = we;	// write enable
			
			cbs.post_tx();		// Post-transfer callback
		end
		endtask
		
		
		// write data to WB bus
		task writecycle(bit [31:0] adr, dat_i);
		begin
			// Set signals on WB bus
			setsignals(1'b1,adr,dat_i);

			@(posedge wbif.clk);
			
			// Sample functional coverage data
			covWbDatIn.sample();

			// Wait for ack from device					
			while (~wbif.wb_ack) @(posedge wbif.clk);

			// End write cycle
			wbif.wb_cyc = 1'b0;
			wbif.wb_stb = 1'b0;

		end
		endtask
		
		
		// read data from WB bus
		task readcycle(bit [31:0] adr);
		begin
			// Set signals on WB bus
			setsignals(1'b0,adr,32'b0);

			@(posedge wbif.clk);
					
			// Wait for ack from device					
			while (~wbif.wb_ack) @(posedge wbif.clk);

			// Fetch data from WB bus
			data_cycle_out[((adr-36)/4)] = wbif.wb_dat_o;

			// End read cycle
			wbif.wb_cyc = 1'b0;
			wbif.wb_stb = 1'b0;

		end
		endtask
		
			
		
	endclass
	
	
		
	
	// *** CHECKER CLASS
	// ***
	class Checker;
		Transaction tr;
	
		mailbox bfm2chk, scb2chk;
	
		function new(mailbox bfm2chk, scb2chk);
			this.bfm2chk = bfm2chk;
			this.scb2chk = scb2chk;
		endfunction

		task run;
			repeat (ROUNDS) begin
	
				// Receive transaction object from BFM. Wait for Scoreboard too
				bfm2chk.get(tr);
				scb2chk.get(tr);
			
				a_output: assert (tr.data_out == tr.data_c_out);
				
				$display("");
				$display("Data in:        %h",tr.data_in);
				$display("Key in:         %h",tr.key);
				$display("Decrypt:        %h",tr.decrypt);
				$display("Reset:          %h",tr.reset);
				$display("BFM data out:   %h",tr.data_out);
				$display("SCB data out:   %h",tr.data_c_out);
				$display("");
	
			end
		endtask	
	endclass
	
	
	
	

	class Environment;

		mailbox gen2bfm;
		mailbox gen2scb;
		mailbox bfm2chk;
		mailbox scb2chk;


		BFM bfm;
		Generator gen;
		Scoreboard scb;
		Checker chk;

		task run;
		begin
			gen2bfm = new;
			bfm2chk = new;
			gen2scb = new;
			scb2chk = new;
			
			gen = new(gen2bfm,gen2scb);
			bfm = new(gen2bfm,bfm2chk);
			scb = new(gen2scb,scb2chk);
			chk = new(bfm2chk,scb2chk);

			$display("*****************************************************");
			$display("* AES WB Test ...                                   *");
			$display("*****************************************************");

			fork
				bfm.run();
				gen.run();
				scb.run();
				chk.run();
			join
	
			$display("*** Finished ***");
		end
		endtask

	endclass
	
	
	
	
	// Define environment
	Environment env;
	
	// Start environment on program start
	initial begin
		env = new;
		env.run();
	end

	
	
	
	
	
endprogram